import os
import logging
from logging.handlers import RotatingFileHandler

from flask import Flask, request

#from . import manager
import manager

from resources import api

__REQUIRED_SETTINGS__ = manager.REQUIRED_SETTINGS[:]
__OPTIONAL_SETTINGS__ = manager.OPTIONAL_SETTINGS[:]

app = Flask(__name__)
if os.environ.get('PYTHONPATH'):
    error_file = os.path.join(os.environ.get('PYTHONPATH'), 'webapp-error.log')
    error_file_handler = RotatingFileHandler(error_file)
    error_file_handler.setLevel(logging.ERROR)
    app.logger.addHandler(error_file_handler)


@app.before_request
def log_request():
    app.logger.info(request)


def load_config(flask_app):
    env_str_keys = set([k for k in list(os.environ.keys())])
    required_keys = set(__REQUIRED_SETTINGS__)
    if required_keys.issubset(env_str_keys):
        config = {k: os.environ[k] for k in __REQUIRED_SETTINGS__}
        config.update({k: os.environ.get(k) for k in __OPTIONAL_SETTINGS__})
        cfg_obj = type('', (object,), config)()
        flask_app.config.from_object(cfg_obj)
    else:
        flask_app.config.from_pyfile('settings.cfg')
    return flask_app

app = load_config(app)

api.init_app(app)


@app.teardown_appcontext
def remove_session(*args, **kwargs):
    if hasattr(app, 'DBSession'):
        app.DBSession.remove()

if __name__ == '__main__':  # pragma: no cover
    app.run(debug=True, processes=1, host='0.0.0.0')
