from ishare.isharedata.lib.DynamicDataSource import DynamicDataSource
import json
import urllib.request, urllib.error, urllib.parse


class JsonDataSource(DynamicDataSource):

    """
    Base JSON WebService Dynamic DataSource.

    Create a subclass to implement a DataSource for a web API that can be
    called via HTTP GET and returns JSON. Subclasses must implement get_url,
    get_fields and parse_data methods.
    """

    __NAME = 'JsonDataSource'
    __DISPLAY_NAME = 'JsonDataSource'
    __IDENTITY = ()

    def __init__(self):
        super(JsonDataSource, self).__init__()

    def query(self, field, value):
        rows = self._get_rows(field, value)
        return DynamicDataSource.to_dataset_xml('DataSource', self.get_fields(rows), rows)

    def preview(self):
        return self.test()

    def test(self):
        rows = self._get_test_rows()
        return DynamicDataSource.to_dataset_xml('DataSource', self.get_fields(rows), rows)

    @property
    def display_name(self):
        settings = dict(self._config.items('Settings'))
        return self._display_name.format(**settings)

    @property
    def fields(self):
        rows = self._get_test_rows()
        return '\n'.join(self.get_fields(rows))

    def get_url(self, field, value):
        """
        The URL to request JSON data from.
        """
        raise NotImplementedError

    def get_fields(self, rows):
        """
        Return a list of fields found in each dict returned by the DataSource

        Arguments:
        rows -- list of dicts returned by parse_data from which fields may
                optionally be determined.
        """
        raise NotImplementedError

    def parse_data(self, data):
        """
        Given a parsed JSON object (data) return a list of dicts containing the
        data that the DataSource should return.
        """
        raise NotImplementedError

    def _get_rows(self, field, value):
        url = self.get_url(field, value)
        r = urllib.request.urlopen(url)
        rows = self.parse_data(json.loads(r.read()))
        return rows

    def _get_test_rows(self):
        value = self._config.get('Test', 'value')
        return self._get_rows(None, value)

    @classmethod
    def copy_keys(cls, lookup, src):
        """Copy keys from src to dest dict renaming keys based on lookup.

        Arguments:
        lookup  -- list of (src_key, dest_key) pairs
        src     -- source dict
        """
        dest = {}
        for k, v in lookup:
            dest[v] = src.get(k)
        return dest
