import logging
import sys
import string
import os
from io import StringIO
from xml.etree.ElementTree import ElementTree as ET
import xml.etree.ElementTree as EMT
from pprint import pprint
from lxml import etree

from ..core import xml2obj


class Conf:
    def __init__(self, settingsFile, logger=None):

        self.logger = logger or logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)

        handler = logging.FileHandler(__name__ + '.log')
        handler.setLevel(logging.DEBUG)

        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)

        self.logger.addHandler(handler)

        self._settings = settingsFile
        self._pathsfile = self._getPathsFilename()
        self._datafile = self._getDataFilename()
        self._paths = open(self._pathsfile, 'r')
        self._data = open(self._datafile, 'r')
        self._configpaths = xml2obj.xml2obj(self._paths)
        self._configdata = xml2obj.xml2obj(self._data)
        self._paths.close()
        self._data.close()
        self.config = self._configdata

        self.config = self._mergeConfigs(self.config, self._configpaths)

        self.xml = self.getXML('xsl/merge.xsl')
        self.iShareMapsXML = self.getiShareMapsXML()

        self.logger.debug('Initialisation complete')


    '''
    Given xpath get an xml fragment
    '''

    def getXMLFragment(self, xpathexp, workspace=None):

        self.logger.debug('In getXMLFragment')
        xml = self.getXML('xsl/merge.xsl')
        s = StringIO(xml)

        self.logger.debug('Workspace = %s', str(workspace))

        if (workspace):

            self.logger.debug('Workspace selected')

            workspace = "'%s'" % workspace

            try:

                fwsxsl = open(self._getWorkspacesXSLPath(), 'r')
                xslt_root = etree.parse(fwsxsl)
                fwsxsl.close()

            except Exception as e:

                self.logger.debug('Problem getting workspaces xsl %s' % e)
                raise

            transform = etree.XSLT(xslt_root)
            isharemapsparam = "'%s'" % self._reverseslash(
                self._getiShareMapsXMLPath())

            self.logger.debug("isharemapsparam=%s" % isharemapsparam)

            try:

                result = transform(etree.XML(xml), workspacename=workspace,
                                   isharemaps=isharemapsparam)

            except Exception as e:

                self.logger.debug(
                    'Problem applying workspaces transformation % s' % e)
                raise

            tree = etree.parse(StringIO(result))

        else:

            tree = etree.parse(s)

        root = tree.getroot()

        try:

            r = root.xpath('/' + xpathexp)[0]

        except:

            self.logger.debug('XPath failed to return a result')
            self.logger.debug('XPath used %s' % xpathexp)
            # self.logger.debug('XML %s' % xml)
            raise

        fragment = etree.tostring(r)

        return '<?xml version="1.0" encoding="utf-8" ?>' + fragment


    '''
    Make sure the fragment is unique, i.e. only one element
    e.g. <?xml version="1.0" encoding="utf-8" ?>
         <URL>http://localhost/featureserver.cgi</URL>
    '''

    def hasChildren(self, xmlstring):
        self.logger.debug('Checking xml %s has no children' % xmlstring)
        root = EMT.fromstring(xmlstring)
        if len(root):
            return True
        else:
            return False

    '''
    Validate the request
    '''

    def validUpdate(self, xpathexp, payloadxmlstring):

        tree = etree.parse(StringIO(self.getXML('merge.xsl')))
        root = tree.getroot()

        # Check payload element name is the same as the last part of the xpath
        payloadtree = etree.parse(StringIO(payloadxmlstring))
        payloadroot = payloadtree.getroot()

        payloadtag = payloadroot.tag

        xpathtop = xpathexp[xpathexp.rfind('/') + 1:]

        if xpathtop != payloadtag:
            self.logger.debug('Update not valid')
            return False
        else:
            self.logger.debug('Update valid')
            return True

    '''
    Validate the delete request
    If iShareMaps does not contain the requested path, reject it
    '''

    def validDelete(self, xpathexp, workspace):
        self.logger.debug("Validating delete")
        settingpath = (
                      "/MapSpur/Workspaces/Workspace[@name='%s']/" % workspace) +\
                      xpathexp
        self.logger.debug("settingpath = %s" % settingpath)

        tree = etree.parse(StringIO(self.getiShareMapsXML()))
        root = tree.getroot()
        valid = False
        if root.xpath(settingpath):
            valid = True
        return valid


    '''
    Delete the xml fragment specified
    '''

    def deleteSetting(self, xpathexp, workspace):
        self.logger.debug("Deleting")
        iShareMapsXML = self.getiShareMapsXML()
        settingpath = ("Workspaces/Workspace[@name='%s']/" % workspace) + xpathexp
        self.logger.debug("settingpath = %s" % settingpath)

        tree = etree.parse(StringIO(iShareMapsXML))
        root = tree.getroot()

        elem = root.find(settingpath)
        elem.getparent().remove(elem)

        # Tidy up using the remove_blank_text option
        untidyxml = etree.tostring(tree, pretty_print=True, xml_declaration=True,
                                   encoding="UTF-8")
        parser = etree.XMLParser(remove_blank_text=True)
        newroot = etree.parse(StringIO(untidyxml), parser).getroot()
        tidyxml = etree.tostring(newroot, pretty_print=True, xml_declaration=True,
                                 encoding="UTF-8")

        self.iShareMapsXML = tidyxml
        self.serialise(workspace)

        return


'''
    Given the xmlstring and the payloadxmlstring, update the xmlstring
    with the value specified in the payloadxmlstring
    '''


def updateSetting(self, xpathexp, payloadxmlstring, workspace=None):
    if workspace:
        iShareMapsXML = self.iShareMapsXML

        # need to prepend the settingpath with xpath specific to iShareMaps.xml
        settingpath = (
                      "/MapSpur/Workspaces/Workspace[@name='%s']/" % workspace) +\
                      xpathexp

        self.logger.debug("settingpath = %s" % settingpath)

        xml = self._createElementIfNotExist(settingpath, iShareMapsXML)

        self.logger.debug('iShareMapsXML = %s' % xml)

        tree = etree.parse(StringIO(xml))
        root = tree.getroot()

        self.logger.debug('Created tree')

        payloadtree = etree.parse(StringIO(payloadxmlstring))
        payloadroot = payloadtree.getroot()

        self.logger.debug('Created payload root')

        #xpathexp = '/' + xpathexp

        xpathexp = settingpath

        self.logger.debug('xpath expression = %s' % xpathexp)

        if root.xpath(xpathexp):
            self.logger.debug('XPath lookup successful')
            eleToChange = root.xpath(xpathexp)[0]
            eleToChange.text = payloadroot.text
            self.logger.debug(
                "Updated XML %s" % etree.tostring(root, pretty_print=True))
            self.iShareMapsXML = etree.tostring(root, pretty_print=True,
                                                xml_declaration=True,
                                                encoding="UTF-8")
            self.serialise(workspace)
        else:
            self.logger.debug('XPath lookup not successful')
            self.logger.debug("Unchanged XML %s" % self.getiShareMapsXML())
        return  #self.getiShareMapsXML()

    else:
        tree = etree.parse(StringIO(self.getXML('merge.xsl')))
        root = tree.getroot()

        payloadtree = etree.parse(StringIO(payloadxmlstring))
        payloadroot = payloadtree.getroot()

        xpathexp = '/' + xpathexp

        self.logger.debug('xpath expression = %s' % xpathexp)

        if root.xpath(xpathexp):
            self.logger.debug('XPath lookup successful')
            eleToChange = root.xpath(xpathexp)[0]
            eleToChange.text = payloadroot.text
            self.logger.debug(
                "Updated XML %s" % etree.tostring(root, pretty_print=True,
                                                  xml_declaration=True,
                                                  encoding="UTF-8"))
            self.xml = etree.tostring(root, pretty_print=True, xml_declaration=True,
                                      encoding="UTF-8")
            self.serialise()
        else:
            self.logger.debug('XPath lookup not successful')
            self.logger.debug("Unchanged XML %s" % self.getXML('merge.xsl'))
    return  # self.getXML('merge.xsl')




def getXML(self, xslStyleSheet):
    '''
    Returns a data.xml and paths.xml merged into a single xml
    '''
    self.logger.debug('In getXML')

    try:
        f = open(self._datafile, 'r')
        s = f.read()
        f.close()
    except IOError:
        self.logger.error('Problem opening data file {0}', self._datafile)
    except:
        self.logger.error('Error getting paths file {0}', sys.exc_info())

    doc = s

    try:
        fmergexsl = open(xslStyleSheet, 'r')
        xslt_root = etree.parse(fmergexsl)
        fmergexsl.close()
    except Exception as e:
        self.logger.error("Problem retrieving merge.xsl file {0}" % e)
        raise

    transform = etree.XSLT(xslt_root)
    pathsparam = "document('%s')" % self._reverseslash(self._getPathsFilename())
    self.logger.debug("pathsparam=%s" % pathsparam)
    result = transform(etree.XML(doc), paths=pathsparam)

    root = etree.XML(str(result))

    # self.logger.debug("xml = %s" % etree.tostring(root,pretty_print=True))

    return etree.tostring(root, pretty_print=True)




def getiShareMapsXML(self):
    '''
    Return the string from iShareMaps.xml
    '''
    ism = ''
    try:
        f = open(self._getiShareMapsXMLPath(), 'r')
        ism = f.read()
    except IOError as e:
        self.logger.debug('Error opening iShareMaps.xml %s' % e)
        raise
    else:
        self.logger.debug('iShareMapsXML loaded.')
        return ism
    finally:
        f.close()


def _getPathsFilename(self):
    tree = ET()
    filename = ''
    try:
        root = tree.parse(self._settings)
        pathsElement = root.find('paths')
        filename = pathsElement.attrib['file']
        f = open(filename, "r")
        f.close()
    except IOError:
        self.logger.error('Problem opening paths file {0}', filename)
    except:
        self.logger.error('Error getting paths file {0}', sys.exc_info())
    self.logger.debug('Paths file name %s' % filename)
    return filename


def _getDataFilename(self):
    tree = ET()
    filename = ''
    try:
        root = tree.parse(self._settings)
        selectedDataConfig = root.find('selectedDataConfig').text
        configList = root.findall('data/config')
        for configElement in configList:
            if configElement.attrib['name'] == selectedDataConfig:
                filename = configElement.attrib['path']
        f = open(filename, "r")
        f.close()
    except IOError:
        self.logger.error('Problem opening data file {0}', filename)
    except:
        self.logger.error('Error getting data file {0}', sys.exc_info())
    self.logger.debug('Data file name %s' % filename)
    return filename





def serialise(self, workspace=None):
    '''
    Take the combined xml and split back into paths.xml and data.xml using
    xsl stylesheets.
    '''
    if workspace:

        self.logger.debug('Validating candidate iShareMaps.xml')

        xsd_doc = etree.parse(sys.path[0] + '/xsd/isharemaps.xsd')
        xsd = etree.XMLSchema(xsd_doc)

        xml = etree.fromstring(self.iShareMapsXML)

        if xsd.validate(xml):
            self.logger.debug('Candidate iShareMaps.xml validates OK')
        else:
            self.logger.debug(
                'Invalid isharemaps.xml, not written %s' % xsd.error_log)
            raise ValidationError('isharemaps.xml invalid')

        self.logger.debug('serialising - write to isharemapsxml')

        try:
            f = open(self._getiShareMapsXMLPath(), 'w')
            f.write(self.iShareMapsXML)
            f.close()
        except Exception as e:
            self.logger.debug('Error writing to iShareMaps.xml %s' % e)

            raise

    else:
        xslt_data_root = etree.parse(
            os.path.abspath("../../shared/config/data.xslt"))
        xslt_paths_root = etree.parse(
            os.path.abspath("../../shared/config/paths.xslt"))

        self.logger.debug('About to parse xml')

        doc = etree.parse(StringIO(self.xml))

        self.logger.debug('xml parse complete')

        transformPaths = etree.XSLT(xslt_paths_root)

        paths = transformPaths(doc)

        transformData = etree.XSLT(xslt_data_root)

        data = transformData(doc)

        xsd_doc = etree.parse(sys.path[0] + '/xsd/paths.xsd')
        xsd = etree.XMLSchema(xsd_doc)

        xml = etree.fromstring(etree.tostring(paths))

        if xsd.validate(xml):
            self.logger.debug('Candidate paths.xml validates OK')
        else:
            self.logger.debug(
                'Invalid paths.xml, not written %s' % xsd.error_log)
            raise ValidationError('Paths.xml invalid')

        xsd_doc = etree.parse(sys.path[0] + '/xsd/data.xsd')
        xsd = etree.XMLSchema(xsd_doc)

        xml = etree.fromstring(etree.tostring(data))

        if xsd.validate(xml):
            self.logger.debug('Candidate data.xml validates OK')
        else:
            self.logger.debug(
                'Invalid data.xml, not written %s' % xsd.error_log)
            raise ValidationError('Data.xml invalid')

        try:

            paths.write(self._getPathsFilename(), pretty_print=True)

        except Exception as e:

            self.logger.error('Problem writing to the paths.xml file %s' % e)
            raise

        try:

            data.write(self._getDataFilename(), pretty_print=True)

        except Exception as e:

            self.logger.error('Problem writing to the data.xml file %s' % e)
            raise

        return False


def _reverseslash(self, stringval):
    return string.replace(stringval, "\\", "/")


def _getiShareMapsXMLPath(self):
    self.logger.debug('In _getiShareMapsXMLPath')

    rootPath = self.config.RootPath

    return rootPath + 'config/iShareMaps.xml'


def _getWorkspacesXSLPath(self):
    self.logger.debug('In _getWorkspacesXSLPath')

    rootPath = self.config.RootPath

    return rootPath + '/xml/Workspaces.xsl'


def _createElementIfNotExist(self, xpathexp, xmlstring):
    self.logger.debug('In _createElementIfNotExist')

    tree = etree.parse(StringIO(xmlstring))
    root = tree.getroot()
    result = root.xpath(xpathexp)

    self.logger.debug('Run xpath query')

    #if the element already exists, just return what is passed in
    if result:
        self.logger.debug('Returning what is passed in')
        return xmlstring

    elements = xpathexp.split("/")

    self.logger.debug('getting root')

    root = etree.XML(xmlstring)
    self.logger.debug('got root')

    missingelements = []
    runningxpath = ''
    elemCount = 0

    self.logger.debug('Iterating through elements in xpath')

    for element in elements:

        if (element != None):
            runningxpath = runningxpath + '/' + element
            if (len(root.xpath(runningxpath)) == 0):
                missingelements.append(runningxpath[1:])
            if (elemCount == 0):
                rootElementName = element

            elemCount = elemCount + 1

    #iterate through the list of missing elements
    self.logger.debug('Iterating through missing elements')

    for element in missingelements:

        self.logger.debug('Element = %s' % element)

        elementsInPath = element.split("/")
        elementToCreate = elementsInPath[-1]

        parent = string.replace(element, elementToCreate, '')
        parent = string.replace(parent, '/MapSpur/', '')[:-1]

        self.logger.debug('Parent = %s' % parent)

        if parent:
            elem = root.find(parent)
            self.logger.debug('Parent element %s ' % elem)
            newelem = etree.Element(elementToCreate)
            self.logger.debug('Element to create %s ' % elementToCreate)
            elem.append(newelem)

    #Tidy up using the remove_blank_text option
    untidyxml = etree.tostring(root, pretty_print=True, xml_declaration=True,
                               encoding="UTF-8")
    parser = etree.XMLParser(remove_blank_text=True)
    newroot = etree.parse(StringIO(untidyxml), parser).getroot()
    tidyxml = etree.tostring(newroot, pretty_print=True, xml_declaration=True,
                             encoding="UTF-8")

    return tidyxml


def _mergeConfigs(self, config, paths):
    #I think there might be a better way of doing this....
    config.External.GdalData = paths.External.GdalData
    config.RootPath = paths.RootPath
    config.External.GdalData = paths.External.GdalData
    config.External.OgrPath = paths.External.OgrPath
    config.External.PythonPath = paths.External.PythonPath
    config.External.ShpTree = paths.External.ShpTree
    config.External.Wkhtmltopdf = paths.External.Wkhtmltopdf
    config.MapServer.ImagePath = paths.MapServer.ImagePath
    config.MapServer.TemplatesPath = paths.MapServer.TemplatesPath
    config.MapServer.URL = paths.MapServer.URL
    config.MapServer.Version = paths.MapServer.Version
    config.MyAlerts.enabled = paths.MyAlerts.enabled

    config.MyAlertsService.enabled = paths.MyAlertsService.enabled
    config.MyAlertsService.AppPath = paths.MyAlertsService.AppPath
    config.MyAlertsService.Debug.enabled = paths.MyAlertsService.Debug.enabled
    config.MyAlertsService.Debug.CacheFile = \
        paths.MyAlertsService.Debug.CacheFile
    config.MyAlertsService.Debug.SnapFile = paths.MyAlertsService.Debug.SnapFile
    config.MyAlertsService.ServicePath = paths.MyAlertsService.ServicePath
    config.MyAlertsService.TransactionWebServiceURL = \
        paths.MyAlertsService.TransactionWebServiceURL
    config.MyAlertsService.UserManagerWebServiceURL = \
        paths.MyAlertsService.UserManagerWebServiceURL
    config.MyAlertsService.UsersWebServiceURL = \
        paths.MyAlertsService.UsersWebServiceURL

    config.Web.URL = paths.Web.URL

    config.WebService.Map = paths.WebService.Map
    config.WebService.PDFMakerURL = paths.WebService.PDFMakerURL
    config.WebService.RSSXSL = paths.WebService.RSSXSL
    config.WebService.URL = paths.WebService.URL

    config.Workflow.ConsoleAppPath = paths.Workflow.ConsoleAppPath
    config.Workflow.SyncAppPath = paths.Workflow.SyncAppPath
    return config


class ValidationError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


if __name__ == '__main__':
    conf = Conf('settings.xml')
    Config = conf.config
    print('GdalData = %s' % Config.External.GdalData)
    print('RootPath = %s' % Config.RootPath)
    print('External = %s' % Config.External)
    print('RootPath = %s' % eval("Config.RootPath"))



