import os

from datetime import datetime as dt
from urllib.parse import urlencode
from urllib.parse import parse_qs, urlsplit, urlunsplit


def get_files_modified(paths=[], relative_to='', suffix='', omit_us=False):
    def f(path):
        full_path = os.path.join(relative_to, path + suffix)
        mtime = os.path.getmtime(full_path)
        modified = dt.fromtimestamp(mtime)
        if omit_us:
            modified = modified.replace(microsecond=0)
        return path, modified
    return list(map(f, paths))


def join_path_or_uri(base, *args):
    path = ''
    if base.startswith('http'):
        base = base.rstrip('/')
        path = '/'.join((base,) + args)
    else:
        path = os.path.join(base, *args)
    return path


def set_qs_parameters(url, param_dict):
    scheme, host, path, qs, fragment = urlsplit(url)
    params = parse_qs(qs)
    for key, value in param_dict.items():
        params[key] = str(value)
    new_qs = urlencode(params, doseq=True)

    return urlunsplit((scheme, host, path, new_qs, fragment))